
/*=============================================================================
                 Copyright (c)  2004 UGS, Inc.
                        Unpublished - All Rights Reserved
=============================================================================*/
/******************************************************************************
*                                                                             *
* DESCRIPTION -                                                               *
*   This program create operation for milling open slots                      *
*                                                                             *
* PROGRAM DESCRIPTION -                                                       *
*   This example requires an open part, solid body with open slots.           *
*   For work this program, need select 1 or more floor fases of the slots     *
*   and asign the plane of MCS location. Program automaticly create setup     * 
*   program group, geometry group, tool and operations for all selected       *                         
******************************************************************************/

/***************************** Include Files **********************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <uf.h>
#include <uf_ui.h>
#include <uf_ui_ont.h>
#include <uf_part.h>
#include <uf_obj.h>
#include <uf_modl.h>
#include <uf_curve.h>
#include <uf_disp.h>
#include <uf_object_types.h>
#include <uf_ui_param.h>
#include <uf_cam.h>
#include <uf_oper.h>
#include <uf_cambnd.h>
#include <uf_camgeom.h>
#include <uf_cam_planes.h>
#include <uf_setup.h>
#include <uf_ncgroup.h>
#include <uf_oper.h>
#include <uf_param.h>
#include <uf_param_indices.h>
#include <uf_lib.h>
#include <uf_eval.h>
#include <uf_vec.h>
#include <uf_csys.h>
#include <uf_mtx.h>
#include <uf_trns.h>
#include <uf_layer.h>
#include <uf_defs.h>
#include <uf_exit.h>
#include <uf_styler.h>
#include <uf_mb.h>
#include <uf_so.h>
#include "mill_slots.h"
//#include "tool_list.h"

#define M_PI 3.14159265358979 
#define MAX_BND 1000
#define LAYER_FOR_ADD_ON_OBJECTS 250
#define MAX_TOOL 100

#define MILL_SLOT_CB_COUNT ( 21 + 1 ) /* Add 1 for the terminator */
#define TOOL_CB_COUNT ( 9 + 1 ) /* Add 1 for the terminator */

static UF_STYLER_callback_info_t MILL_SLOT_cbs[MILL_SLOT_CB_COUNT] = 
{
 {UF_STYLER_DIALOG_INDEX, UF_STYLER_CONSTRUCTOR_CB  , 0, MILL_SLOT_constructor_cb},
 {UF_STYLER_DIALOG_INDEX, UF_STYLER_OK_CB           , 1, MILL_SLOT_ok_cb},
 {UF_STYLER_DIALOG_INDEX, UF_STYLER_APPLY_CB        , 1, MILL_SLOT_apply_cb},
 {UF_STYLER_DIALOG_INDEX, UF_STYLER_CANCEL_CB       , 0, MILL_SLOT_cancel_cb},
 {MILL_SLOT_TOOL_TYPE   , UF_STYLER_ACTIVATE_CB     , 0, MILL_SLOT_Tool_Creation_Type},
 {MILL_SLOT_PERCENT_OF_TROCHOIDA_WIDTH, UF_STYLER_ACTIVATE_CB, 0, MILL_SLOT_percent_of_trochoida_width},
 {MILL_SLOT_MIN_OF_TROCHOIDA_WIDTH, UF_STYLER_ACTIVATE_CB, 0, MILL_SLOT_Min_of_trochoida_width},
 {MILL_SLOT_DIAMETER_1  , UF_STYLER_ACTIVATE_CB     , 0, MILL_SLOT_tool_diameter},
 {MILL_SLOT_LENGTH_2    , UF_STYLER_ACTIVATE_CB     , 0, MILL_SLOT_tool_length},
 {MILL_SLOT_FLUTE_LENGTH_3, UF_STYLER_ACTIVATE_CB   , 0, MILL_SLOT_tool_flute_length},
 {MILL_SLOT_NUMBER_4    , UF_STYLER_ACTIVATE_CB     , 0, MILL_SLOT_tool_number},
 {MILL_SLOT_ADJUST_REGISTER_5, UF_STYLER_ACTIVATE_CB, 0, MILL_SLOT_tool_adjust_register},
 {MILL_SLOT_CUTCOM_REGISTER_8, UF_STYLER_ACTIVATE_CB, 0, MILL_SLOT_tool_cutcom_register},
 {MILL_SLOT_CUT_TYPE_9  , UF_STYLER_ACTIVATE_CB     , 0, MILL_SLOT_operation_cut_type},
 {MILL_SLOT_TROCHOIDA_DIRECTION, UF_STYLER_ACTIVATE_CB, 0, MILL_SLOT_Trochoida_direction},
 {MILL_SLOT_REAL_12     , UF_STYLER_ACTIVATE_CB     , 0, MILL_SLOT_operation_min_clearance},
 {MILL_SLOT_DEPTH_PER_CUT_12, UF_STYLER_ACTIVATE_CB , 0, MILL_SLOT_operation_depth_per_cut},
 {MILL_SLOT_PART_STOCK_15, UF_STYLER_ACTIVATE_CB    , 0, MILL_SLOT_operation_part_stock},
 {MILL_SLOT_TROCHOIDA_STEP_16, UF_STYLER_ACTIVATE_CB, 0, MILL_SLOT_operation_trochoida_step},
 {MILL_SLOT_REAL_16     , UF_STYLER_ACTIVATE_CB     , 0, MILL_SLOT_operation_trochoida_start_angle},
 {MILL_SLOT_CREATE_OPERATIONS, UF_STYLER_ACTIVATE_CB, 0, MILL_SLOT_Operations_types},
 {UF_STYLER_NULL_OBJECT, UF_STYLER_NO_CB, 0, 0 }
};

static UF_STYLER_callback_info_t TOOL_cbs[TOOL_CB_COUNT] = 
{
 {UF_STYLER_DIALOG_INDEX, UF_STYLER_CONSTRUCTOR_CB  , 0, TOOL_constructor_cb},
 {UF_STYLER_DIALOG_INDEX, UF_STYLER_OK_CB           , 0, TOOL_ok_cb},
 {UF_STYLER_DIALOG_INDEX, UF_STYLER_APPLY_CB        , 0, TOOL_apply_cb},
 {UF_STYLER_DIALOG_INDEX, UF_STYLER_CANCEL_CB       , 0, TOOL_cancel_cb},
 {TOOL_LIST_0           , UF_STYLER_ACTIVATE_CB     , 0, TOOL_tool_list_activate},
 {TOOL_LIST_0           , UF_STYLER_DOUBLE_CLICK_CB , 0, TOOL_tool_list_click},
 {TOOL_DIAMETER         , UF_STYLER_ACTIVATE_CB     , 0, TOOL_diameter},
 {TOOL_FLUTELENGTH      , UF_STYLER_ACTIVATE_CB     , 0, TOOL_FluteLength},
 {TOOL_MATERIALDES      , UF_STYLER_ACTIVATE_CB     , 0, TOOL_MaterialDes},
 {UF_STYLER_NULL_OBJECT, UF_STYLER_NO_CB, 0, 0 }
};

static UF_MB_styler_actions_t actions[] = {
    { "mill_slot.dlg",  NULL,   MILL_SLOT_cbs,  UF_MB_STYLER_IS_NOT_TOP },
    { NULL,  NULL,  NULL,  0 } /* This is a NULL terminated list */
};

static UF_MB_styler_actions_t actions1[] = {
    { "tool_list.dlg",  NULL,   TOOL_cbs,  UF_MB_STYLER_IS_NOT_TOP },
    { NULL,  NULL,  NULL,  0 } /* This is a NULL terminated list */
};

/*******************************************************************/
static UF_CAMBND_boundary_data_p_t create_and_init_bnd_data
(
    UF_CAM_boundary_type_t   type,
    int                      plane_type,
    double                   *origin,
    double                   *matrix,
    UF_CAM_material_side_t   material_side,
    int                      ignore_holes,
    int                      ignore_islands,
    int                      ignore_chamfers,
    UF_CAMBND_app_data_p_t   app_data
);

static UF_CAMBND_boundary_data_p_t create_and_init_finish_bnd_data
(
    UF_CAM_boundary_type_t   type,
    int                      plane_type,
    double                   *origin,
    double                   *matrix,
    UF_CAM_material_side_t   material_side,
    int                      ignore_holes,
    int                      ignore_islands,
    int                      ignore_chamfers,
    UF_CAMBND_app_data_p_t   app_data
);

static UF_CAMBND_app_data_p_t create_and_init_bnd_app_data(void);
static UF_CAMBND_app_data_p_t create_and_init_finish_bnd_app_data(void);
static int init_proc(UF_UI_selection_p_t select, void *user_data);
void get_error(int irc);
static void set_operation_param(tag_t operation, tag_t floor_face, 
								tag_t from_tag, int from_gohome_flag);
static void set_operation_param_finish(tag_t operation, tag_t floor_face, 
									   tag_t from_tag, int from_gohome_flag);
static void init_camgeom_app_data(UF_CAMGEOM_app_data_p_t app_data);
void change_tool_parameters(double tool_diameter, double tool_heigth,
		double tool_flute_ln, int tool_number,
		int tool_adj_reg, int tool_cutcom_reg, tag_t new_tool);
void change_tool_parameters_auto(int tool_number, int tool_adj_reg, 
							int tool_cutcom_reg, tag_t new_tool);
void unhighlight_objects(int count, tag_p_t object_p_t);
void print_to_listing_windows(char str[256]);
static void ask_edge_end(tag_t edge, double point[3]);
static void ask_edge_start(tag_t edge, double point[3]);
tag_t create_trochoida(double tolerance, double start_point[3], 
		double distances[3], double csys[9], double tool_diameter,
		double stock, double step, double start_angle, 
		tag_t *edges_list, int count_edge, tag_t start_edge,
		double directions[3][3]);
static void display_temporary_point(double *coords);
tag_t open_edge(tag_t *edges_array, int number_edges, 
				double tolerance);
tag_t create_boundary_curve(tag_t face_objects_sel, double origin[3],
							double tool_diameter,
							double stock, double step, 
							double start_angle, int cut_type);
void create_boundary_edges(tag_t face_objects_sel, double origin[3]);
void create_boundary(tag_t boundary_curve, double origin[3], 
					 tag_t face_id, tag_t cam_oper);
void create_finish_boundary(double origin[3], tag_t face_id, 
							tag_t finish_operation);
static void do_mill_slots_api(void);
static void sort_end_to_end(tag_t *curves, int n_curves);
double ask_min_width(tag_t *face_objects, int face_count);
double calculate_tool_diameter(double min_width, 
							   double stock, int cut_type);
int queryToolDiameter(double tool_diam, double cor_radius, 
					  double fl_length);
extern int tool_list_menu(int *response);
static void traverseGroups(tag_t group);

/* Global variables */
/* Parameters input from menu */
/*===================================================*/
tag_t *bnd_edges;
char *returned_tool_list[30+1];
double tool_diameter = 10.0;
double tool_heigth = 40.0;
double tool_flute_ln = 20.0;
double min_clr = 3.0;
double depth_per_cut = 2.0;
double stock = 0.5;
double step = 1.0;
double start_angle = 180;
double trochoida_width = 0.7;
double trochoida_width_min = 0.5;
double tool_diam_array[MAX_TOOL];
double tool_flute_ln_array[MAX_TOOL];
int num_bnd_curves = 0;
int operetion_types = 0;
int tool_creation_method = 0;
int trochoida_dir = 0;
int tool_number = 1;
int tool_adj_reg = 1;
int	tool_cutcom_reg = 1;
int cut_type = 1; /* 0 - line 1 - trochoida*/
int number_of_tool_ret;
int index_tool = 0;
char *material_description[256];
int operation_exist;

/*===================================================*/

/*******************************************************************/
static void do_mill_slots_api(void)
{
	logical generated;
    int i = 0, j, irc, grip_arg_count = 1;
	int count_oper, entity_count = 0, response;
	int face_count = 0, mode = 0, display = 0;
	int from_gohome_flag;
//	int is_equal_origin, is_equal_normal;
	char str[256] = "", **type_name;
	char cue_face[] = "Select floor faces of slots";
	char title[] = "Face selection";
	char grip_exe[256], *translation;
//	char oper_name[UF_OPER_MAX_NAME_LEN + 1];
	char tol_libref[30+1];
	tag_t new_oper, part_tag, new_tool, mct_group, new_oper_finish;
	tag_t boundary_curve, mcs_csys, setup_tag, geom_grp; 
	tag_t body_obj_id, matrix_mcs_csys, from_point_tag;
	tag_t group_geom = NULL_TAG, group_bnd = NULL_TAG;
	tag_t prog_grp = NULL_TAG, nc_prog_grp = NULL_TAG;
	tag_t plane_eid = NULL_TAG, group_mcs = NULL_TAG;
	tag_t nc_prog_grp_fin = NULL_TAG;
	tag_p_t face_objects;
	double origin[3], cl_plane_origin[3], cl_plane_normal[3];
	double start_point_body[3], directions_body[3][3], distances_body[3];
	double orientation[9], offset[6] = {0.0,0.0,0.0,0.0,0.0,0.0};
	double from_point_coord[3], from_distance = 100.0;
	double min_width, cor_radius = 0.0, fl_length = 20.0;
//	double origin_exist[3]; 
//	double normal_exist[3];
	double modl_tolerance;
//	UF_args_t grip_arg_list[2];
	UF_CAM_geom_type_t geometry_type = UF_CAM_part;
	UF_CAMGEOM_app_data_p_t app_data_list_1[MAX_BND];
//	UF_PARAM_clrplane_status_t cl_status;

/* Ask display part tag. */
    part_tag = UF_PART_ask_display_part();

	irc = UF_translate_variable("UGII_USER_DIR", &translation);
	if(irc || translation == NULL) 
	{
		uc1601("System variable UGII_USER_DIR not set !!!\n Set UGII_USER_DIR variable and restart UG !!!\n", TRUE);
		return;
	}
	strcpy(grip_exe, translation);
	strcat(grip_exe, "\\application\\set_floor_plane.grx");

/* Select floor fases of slots */
	do
	{
		irc = UF_UI_select_with_class_dialog(cue_face, title, 
			UF_UI_SEL_SCOPE_ANY_IN_ASSEMBLY, init_proc, NULL, 
			&response, &face_count, &face_objects);
		if(irc) get_error(irc);
		if(response != UF_UI_OK || face_count == 0)
		{
			uc1601("Fases not selected",TRUE);
		}
	} while(face_count == 0);
	unhighlight_objects(face_count, face_objects);
/* Ask modeling tolerance */
	UF_MODL_ask_distance_tolerance(&modl_tolerance);

/* If Tool Creation Method = Automatic, ask tool from database */
	if(tool_creation_method == 1)
	{
/* Ask minimun slots width */
		min_width = ask_min_width(face_objects, face_count);
/* Calculate tool diameter */
		tool_diameter = calculate_tool_diameter(min_width, stock, cut_type);
	}
/* Ask body tag by face */
	irc = UF_MODL_ask_face_body(face_objects[0], &body_obj_id);
	if(irc) get_error(irc);
	
/* Create MCS */
	irc = UF_UI_specify_plane("Select Plane of MCS location", 
		&mode, display, &response, orientation, origin, &plane_eid);
	if(irc) get_error(irc);
	if(response != 3 && plane_eid == NULL_TAG)
	{
		uc1601("Plane not selected!\nSelect plane of MCS location",TRUE);
	} else if (response == 1 || response == 2) 
		return;
	irc = UF_CSYS_create_matrix(orientation, &matrix_mcs_csys);
	if(irc) get_error(irc);
	irc = UF_CSYS_create_csys(origin, matrix_mcs_csys, &mcs_csys);
	if(irc) get_error(irc);
	irc = UF_CSYS_set_wcs(mcs_csys);
	if(irc) get_error(irc);

	irc = UF_CAM_init_session();
	if(irc) get_error(irc);
	irc = UF_CAM_opt_ask_types(&count_oper, &type_name);
	if(irc) get_error(irc);
/* Ask CAM setup. If not exist, create */
	irc = UF_SETUP_ask_setup(&setup_tag);
	if(irc) get_error(irc);
	if(setup_tag == NULL_TAG)
	{
		irc = UF_SETUP_create(part_tag, "mill_open_slots");
		if(irc) get_error(irc);
		irc = UF_SETUP_ask_setup(&setup_tag);
		if(irc) get_error(irc);
	}
/* Create nc program group */
	UF_SETUP_ask_program_root(setup_tag, &prog_grp);
	irc = UF_OBJ_cycle_by_name("SLOTS_PROG", &nc_prog_grp);
	if(irc) get_error(irc);
	UF_NCGROUP_accept_member(prog_grp, nc_prog_grp);
	irc = UF_OBJ_cycle_by_name("SLOTS_PROG_FINICH", &nc_prog_grp_fin);
	if(irc) get_error(irc);
	UF_NCGROUP_accept_member(prog_grp, nc_prog_grp_fin);
/* Create mcs group */
	UF_SETUP_ask_geom_root(setup_tag, &geom_grp);
//	irc = UF_OBJ_cycle_by_name("MCS_SLOTS", &group_mcs);
//	if(irc) get_error(irc);
	UF_NCGEOM_create("mill_open_slots", "MCS", &group_mcs);
	UF_NCGROUP_accept_member(geom_grp, group_mcs);
/* Set Clearance Plane Parameters */	
	irc = UF_MODL_ask_bounding_box_aligned(body_obj_id, mcs_csys, FALSE,
			start_point_body, directions_body, distances_body);
	if(irc) get_error(irc);
	for(i = 0; i <= 2; i++)
	{
		cl_plane_origin[i] = start_point_body[i];
		for(j = 0; j <= 2; j++)
		{
			cl_plane_origin[i] += directions_body[j][i]*distances_body[j];
		}
	}
	from_point_coord[0] = origin[0]+from_distance*directions_body[2][0];
	from_point_coord[1] = origin[1]+from_distance*directions_body[2][1];
	from_point_coord[2] = origin[2]+from_distance*directions_body[2][2];
	irc = UF_CURVE_create_point(from_point_coord, &from_point_tag);
	if(irc) get_error(irc);

	cl_plane_origin[0] = cl_plane_origin[0]+min_clr*directions_body[2][0];
	cl_plane_origin[1] = cl_plane_origin[1]+min_clr*directions_body[2][1];
	cl_plane_origin[2] = cl_plane_origin[2]+min_clr*directions_body[2][2];
	cl_plane_normal[0] = directions_body[2][0];
	cl_plane_normal[1] = directions_body[2][1];
	cl_plane_normal[2] = directions_body[2][2];

/*        
          */	
/*
	traverseGroups(geom_grp);
	if(operation_exist == 1)
	{
sprintf(str, "Operation exist = %d\n", operation_exist);
uc1601(str, TRUE);
	}
*/
/*  */
/* If clearance plane exist, check the origin and normal */
/*
	irc = UF_CAM_ask_clear_plane_status(group_mcs, &cl_status);
	if(irc) get_error(irc);
	if(cl_status == UF_PARAM_clrplane_define_and_active)
	{
		irc = UF_CAM_ask_clear_plane_data(group_mcs, 
			origin_exist, normal_exist);
		if(irc) get_error(irc);

		UF_VEC3_is_equal(origin_exist, 
			cl_plane_origin, modl_tolerance, &is_equal_origin);
		UF_VEC3_is_equal(origin_exist, 
			cl_plane_origin, modl_tolerance, &is_equal_normal);
		if(is_equal_origin == 0 || is_equal_normal == 0)
uc1601("Not same cl plane", TRUE);
	}
*/
/* Add clearance plane in MCS group */
	UF_CAM_set_clear_plane_status(group_mcs, 
		UF_PARAM_clrplane_define_and_active);
	UF_CAM_set_clear_plane_usage(group_mcs,
		UF_PARAM_clrplane_usage_at_operation_start_and_end);
	UF_CAM_set_clear_plane_data(group_mcs, cl_plane_origin, cl_plane_normal);
/* Create geometry group, rename to 'SLOTS_GEOMETRY',
append part and blank geometry */
//	irc = UF_OBJ_cycle_by_name("SLOTS_GEOMETRY", &group_geom);
//	if(irc) get_error(irc);
	UF_NCGEOM_create("mill_open_slots", "SLOTS_GEOMETRY", &group_geom);
	UF_NCGROUP_accept_member(group_mcs, group_geom);
/* Allocate the memory for the application data of an item. */
	app_data_list_1[0] = (UF_CAMGEOM_app_data_p_t)
		UF_allocate_memory(sizeof(UF_CAMGEOM_app_data_t), &irc);
/* Initialize the application data. */
	init_camgeom_app_data(app_data_list_1[0]);
/* Add solid body in geometry group as Part geometry */
	UF_CAMGEOM_append_items(group_geom, geometry_type, 1, 
		&body_obj_id, app_data_list_1);
/* Add auto blank geometry */
	UF_CAM_set_auto_blank(group_geom, UF_CAM_auto_block_type, offset);
/* Create boundary and rename it to 'SLOTS_BND' 
This object not using, but need to create boundary
inside in operation	*/
//	irc = UF_OBJ_cycle_by_name("SLOTS_BND", &group_bnd);
//	if(irc) get_error(irc);
	irc = UF_NCGEOM_create("mill_open_slots", "SLOTS_BND", &group_bnd);
	if(irc) get_error(irc);
	irc = UF_NCGROUP_accept_member(group_geom, group_bnd); 
	if(irc) get_error(irc);
/* Creating and call tool */
	if(tool_creation_method == 1)
	{
/* Call tool from database */
		queryToolDiameter(tool_diameter, cor_radius, fl_length);
		irc = tool_list_menu(&response);
		if(irc) get_error(irc);
		strcpy(tol_libref, returned_tool_list[index_tool]);
		UF_CUTTER_retrieve(tol_libref, &new_tool);
		UF_SETUP_ask_mct_root(setup_tag, &mct_group);
		UF_NCGROUP_accept_member(mct_group, new_tool);
		change_tool_parameters_auto(tool_number, tool_adj_reg, 
			tool_cutcom_reg, new_tool);
	}
	else
	{
/* Create tool */
		UF_SETUP_ask_mct_root(setup_tag, &mct_group);
		UF_CUTTER_create("mill_open_slots", "MILL", &new_tool);
		UF_NCGROUP_accept_member(mct_group, new_tool);
/* Change tool parameters */
		change_tool_parameters(tool_diameter, tool_heigth, tool_flute_ln, 
				tool_number, tool_adj_reg, tool_cutcom_reg, new_tool);
	}
/* Check names of existing operations */

	if(operetion_types == 0 || operetion_types == 1)
	{
		for(i=0; i<face_count; i++)
		{
			UF_OPER_create("mill_open_slots", "MILL_SLOT", &new_oper);
			UF_NCGROUP_accept_member(group_bnd, new_oper);
			if(i == 0) from_gohome_flag = 1;
			else if(i == face_count-1) from_gohome_flag = 2;
			else from_gohome_flag = 0;
			set_operation_param(new_oper, face_objects[i], from_point_tag, 
								from_gohome_flag);
/* Assing tool to operation*/
			UF_NCGROUP_accept_member(new_tool, new_oper);
/* Assing operation to program group */
			UF_NCGROUP_accept_member(nc_prog_grp, new_oper);
/* Create boundary curve (trochoida or line) */
			boundary_curve = create_boundary_curve(face_objects[i], 
				origin, tool_diameter, stock, step, start_angle, cut_type);
/* Create boundary */
			create_boundary(boundary_curve, origin, 
				face_objects[i], new_oper);
/* Generate operation */
			irc = UF_PARAM_generate(new_oper, &generated);
			if(irc) get_error(irc);
/* Set layer for boundary curves */
			irc = UF_OBJ_set_layer(boundary_curve, LAYER_FOR_ADD_ON_OBJECTS);
			if(irc) get_error(irc);
		}
	}
	if(operetion_types == 0 || operetion_types == 2)
	{
		for(i=0; i<face_count; i++)
		{
			UF_OPER_create("mill_open_slots", "MILL_SLOT_FINISH", &new_oper_finish);
			UF_NCGROUP_accept_member(group_bnd, new_oper_finish);
			if(i == 0) from_gohome_flag = 1;
			else if(i == face_count-1) from_gohome_flag = 2;
			else from_gohome_flag = 0;
			set_operation_param_finish(new_oper_finish, face_objects[i],
										from_point_tag, from_gohome_flag);
/* Assing tool to operation*/
			UF_NCGROUP_accept_member(new_tool, new_oper_finish);
/* Assing operation to program group */
			UF_NCGROUP_accept_member(nc_prog_grp_fin, new_oper_finish);
/* Create boundary curve (trochoida or line) */
			create_boundary_edges(face_objects[i], origin);
/* Create boundary to finish operation */
			create_finish_boundary(origin, 
				face_objects[i], new_oper_finish);
/* Generate operation */
			irc = UF_PARAM_generate(new_oper_finish, &generated);
			if(irc) get_error(irc);
		}
	}
/* Set layer LAYER_FOR_ADD_ON_OBJECTS to invizible state */
	irc = UF_LAYER_set_status(LAYER_FOR_ADD_ON_OBJECTS,	UF_LAYER_INACTIVE_LAYER);
	if(irc) get_error(irc);
/* Move add_on geometry to invizible layer */
	irc = UF_OBJ_set_layer(from_point_tag, LAYER_FOR_ADD_ON_OBJECTS);
	if(irc) get_error(irc);
	irc = UF_OBJ_set_layer(mcs_csys, LAYER_FOR_ADD_ON_OBJECTS);
	if(irc) get_error(irc);

	free(bnd_edges);
//	free(returned_tool_list);
	UF_free(face_objects);
	return;
}

static UF_CAMBND_boundary_data_p_t create_and_init_bnd_data
	(UF_CAM_boundary_type_t type, int plane_type, double *origin,
    double *matrix, UF_CAM_material_side_t material_side,
    int ignore_holes, int ignore_islands, int ignore_chamfers,
    UF_CAMBND_app_data_p_t app_data)
{
    UF_CAMBND_boundary_data_p_t bnd_data;
    int i, err_code;

	type = UF_CAM_boundary_type_open;
	material_side = UF_CAM_material_side_undefined;

    bnd_data = (UF_CAMBND_boundary_data_p_t)UF_allocate_memory(
				sizeof(UF_CAMBND_boundary_data_t), &err_code);
    bnd_data->boundary_type = type;
    bnd_data->plane_type = plane_type;
    for (i=0; i<3; i++)
        bnd_data->origin[i] = origin[i];
    for (i=0; i<9; i++)
        bnd_data->matrix[i] = matrix[i];
    bnd_data->material_side = material_side;
    bnd_data->ignore_holes = ignore_holes;
    bnd_data->ignore_islands = ignore_islands; 
    bnd_data->ignore_chamfers = ignore_chamfers;    
    bnd_data->app_data = app_data;

    return(bnd_data);
}

static UF_CAMBND_boundary_data_p_t create_and_init_finish_bnd_data
	(UF_CAM_boundary_type_t type, int plane_type, double *origin,
    double *matrix, UF_CAM_material_side_t material_side,
    int ignore_holes, int ignore_islands, int ignore_chamfers,
    UF_CAMBND_app_data_p_t app_data)
{
    UF_CAMBND_boundary_data_p_t bnd_data;
    int i, err_code;

	type = UF_CAM_boundary_type_open;
	material_side = UF_CAM_material_side_undefined;

    bnd_data = (UF_CAMBND_boundary_data_p_t)UF_allocate_memory(
				sizeof(UF_CAMBND_boundary_data_t), &err_code);
    bnd_data->boundary_type = type;
    bnd_data->plane_type = plane_type;
    for (i=0; i<3; i++)
        bnd_data->origin[i] = origin[i];
    for (i=0; i<9; i++)
        bnd_data->matrix[i] = matrix[i];
    bnd_data->material_side = material_side;
    bnd_data->ignore_holes = ignore_holes;
    bnd_data->ignore_islands = ignore_islands; 
    bnd_data->ignore_chamfers = ignore_chamfers;    
    bnd_data->app_data = app_data;

    return(bnd_data);
}

static UF_CAMBND_app_data_p_t create_and_init_bnd_app_data(void)
{
    UF_CAMBND_app_data_p_t app_data; 
    int err_code;

    app_data = (UF_CAMBND_app_data_p_t) UF_allocate_memory(
					sizeof(UF_CAMBND_app_data_t), &err_code);
    if(err_code) return(NULL);
/* Set flags. */
    app_data->has_stock = 0;
    app_data->has_tolerances = 0;
    app_data->has_feedrate = 0;
    app_data->has_blank_distance = 0;
    app_data->has_tool_position = 1;
/* Set values. */
    app_data->stock = 0.1;
    app_data->tolerances[0] = 0.0;
    app_data->tolerances[1] = 0.01;
    app_data->feedrate_unit = UF_CAM_feedrate_unit_per_minute;
    app_data->feedrate_value = 200.0;
    app_data->blank_distance = 0.0;
    app_data->tool_position = UF_CAM_tool_position_on;
    return(app_data);
}

static UF_CAMBND_app_data_p_t create_and_init_finish_bnd_app_data(void)
{
    UF_CAMBND_app_data_p_t app_data; 
    int err_code;

    app_data = (UF_CAMBND_app_data_p_t) UF_allocate_memory(
					sizeof(UF_CAMBND_app_data_t), &err_code);
    if(err_code) return(NULL);
/* Set flags. */
    app_data->has_stock = 0;
    app_data->has_tolerances = 0;
    app_data->has_feedrate = 0;
    app_data->has_blank_distance = 0;
    app_data->has_tool_position = 1;
/* Set values. */
    app_data->stock = 0.1;
    app_data->tolerances[0] = 0.0;
    app_data->tolerances[1] = 0.01;
    app_data->feedrate_unit = UF_CAM_feedrate_unit_per_minute;
    app_data->feedrate_value = 200.0;
    app_data->blank_distance = 0.0;
    app_data->tool_position = UF_CAM_tool_position_tanto;
    return(app_data);
}

/* Main function */
void ufusr(char *param, int *retcode, int param_len)
{
	int irc, response;

    UF_initialize();
    irc = UF_STYLER_create_dialog("mill_slots.dlg", MILL_SLOT_cbs,
           MILL_SLOT_CB_COUNT, NULL, &response);
	if(irc) get_error(irc);
    UF_terminate();
}

static void init_camgeom_app_data(UF_CAMGEOM_app_data_p_t app_data)
{
	if(app_data)
	{
/* Set flags. */
		app_data->has_stock = 0;
		app_data->has_cut_stock = 0;
		app_data->has_tolerances = 0;
		app_data->has_feedrate = 0;
		app_data->has_offset = 0;
		app_data->has_avoidance_type = 0;

/* Set values. */
		app_data->stock = 0.001 ;
        app_data->cut_stock[0] = 0.0 ;
        app_data->cut_stock[1] = 0.0 ;
        app_data->cut_stock[2] = 0.0 ;
        app_data->tolerances[0] = 0.000 ;
        app_data->tolerances[1] = 0.001 ;
        app_data->feedrate_unit = UF_CAM_feedrate_unit_per_minute ;
        app_data->feedrate_value = 250.0 ;
        app_data->offset = 0.0 ;
        app_data->avoidance_type =  UF_CAM_avoidance_type_warning;
    }
}
/* Create boundary curve: if cut_type = 0 return line, 
if cut_type = 1 return trochoida */
tag_t create_boundary_curve(tag_t face_objects_sel, 
							double origin[3], double tool_diameter,
							double stock, double step, 
							double start_angle, int cut_type)
{
	char str[256] = "";
	int edge_count, face_count;
	int i, j, irc, edge_type, norm_dir, type;
	tag_t line = NULL_TAG, edge_object, max_edge, csys_id;
	tag_t matrix_id, trochoida_id = 0;
	tag_t *list_edges, start_edge, face_object;
	double arc_length, max_length = 0, tolerance, magnitude;
	double start_point[3], end_point[3], vector_dir[3], unit_vec1[3];
	double distances[3], tmp_p[3] = {0.0,0.0,0.0}, cross_product[3];
	double mtx[9], directions[3][3], dir[3], z_max;
	double min_dist, min_dist_1, pt_on_ent1[3], pt_on_ent2[3], guess1[3];
	double start_point_max[3], directions_max[3][3], distances_max[3] = {0.0, 0.0, 0.0};
	double point[3], box[6], radius, rad_data;

	UF_CURVE_line_t line_coords;
	uf_list_p_t edge_list, face_list;

	UF_MODL_ask_distance_tolerance(&tolerance);

	irc = UF_MODL_ask_face_data(face_objects_sel, &type, point, 
				dir, box, &radius, &rad_data, &norm_dir);
	if(irc) get_error(irc);

	irc = UF_MODL_ask_face_edges(face_objects_sel, &edge_list);
	if(irc) get_error(irc);
	irc = UF_MODL_ask_list_count(edge_list, &edge_count);
	if(irc) get_error(irc);
	list_edges = (tag_t *)malloc(sizeof(tag_t));
	j = 1;
	for(i = 0; i < edge_count; i++)
	{
		irc = UF_MODL_ask_list_item(edge_list, i, &edge_object);
		if(irc) get_error(irc);

		list_edges[i] = edge_object;
		list_edges = (tag_t *)realloc(list_edges, sizeof(tag_t)*(j+1));
		j++;
		irc = UF_MODL_ask_edge_type(edge_object, &edge_type);
		if(irc) get_error(irc);
		if(edge_type == UF_MODL_LINEAR_EDGE)
		{
			irc = UF_CURVE_ask_arc_length(edge_object, 0.0, 1.0, 
				UF_MODL_UNITS_PART, &arc_length);
			if(irc) get_error(irc);
			if(arc_length > max_length) 
			{
				max_length = arc_length;
				max_edge = edge_object;
			}
		}
	}
	ask_edge_start(max_edge, start_point);
	ask_edge_end(max_edge, end_point);
	start_edge = open_edge(list_edges, edge_count, tolerance);

	UF_MODL_ask_minimum_dist(start_edge, NULL_TAG, 0, guess1, 
		1, start_point, &min_dist, pt_on_ent1, pt_on_ent2);
	UF_MODL_ask_minimum_dist(start_edge, NULL_TAG, 0, guess1, 
		1, end_point, &min_dist_1, pt_on_ent1, pt_on_ent2);

	if(min_dist > min_dist_1)
	{
		for(i = 0; i <= 2; i++)
		{
			tmp_p[i] = end_point[i];
			end_point[i] = start_point[i];
			start_point[i] = tmp_p[i];
		}
	}
	vector_dir[0] = end_point[0]-start_point[0];
	vector_dir[1] = end_point[1]-start_point[1];
	vector_dir[2] = end_point[2]-start_point[2];

	irc = UF_VEC3_unitize(vector_dir, tolerance, &magnitude, unit_vec1);
	if(irc) get_error(irc);

	UF_VEC3_negate(dir, dir);

	UF_VEC3_cross(vector_dir, dir, cross_product);

	irc = UF_MTX3_initialize(vector_dir, cross_product, mtx);
	if(irc) get_error(irc);

	irc = UF_CSYS_create_matrix(mtx, &matrix_id);
	if(irc) get_error(irc);
	irc = UF_CSYS_create_csys(start_point, matrix_id, &csys_id);
	if(irc) get_error(irc);

	irc = UF_MODL_ask_bounding_box_aligned(face_objects_sel, csys_id, FALSE, 
			start_point, directions, distances);
	if(irc) get_error(irc);

	for(i = 0; i <= 2; i++)
	{
		end_point[i] = start_point[i];
		for(j = 0; j <= 2; j++)
		{
			end_point[i] += directions[j][i]*distances[j];
		}
	}
	z_max = -1.0;
	for(i = 0; i < edge_count; i++)
	{
		irc = UF_MODL_ask_list_item(edge_list, i, &edge_object);
		if(irc) get_error(irc);

		if(edge_object != start_edge)
		{
			irc = UF_MODL_ask_edge_faces(edge_object, &face_list);
			if(irc) get_error(irc);
			irc = UF_MODL_ask_list_count(face_list, &face_count);
			if(irc) get_error(irc);
			for(j = 0; j < face_count; j++)
			{
				irc = UF_MODL_ask_list_item(face_list, j, &face_object);
				if(irc) get_error(irc);
				if(face_object != face_objects_sel)
				{
					irc = UF_MODL_ask_bounding_box_aligned(
						face_object, csys_id, FALSE, start_point_max, 
						directions_max, distances_max);
					if(irc) get_error(irc);
				}
				if(distances_max[2] > z_max) z_max = distances_max[2];
			}
		}
	}

	if(cut_type == 0)
	{
		line_coords.start_point[0] = start_point[0]+distances[1]/2*directions[1][0];
		line_coords.start_point[1] = start_point[1]+distances[1]/2*directions[1][1];
		line_coords.start_point[2] = start_point[2]+distances[1]/2*directions[1][2];
		line_coords.end_point[0] = end_point[0]-distances[1]/2*directions[1][0];
		line_coords.end_point[1] = end_point[1]-distances[1]/2*directions[1][1];
		line_coords.end_point[2] = end_point[2]-distances[1]/2*directions[1][2];
		irc = UF_CURVE_create_line(&line_coords, &line);
		if(irc) get_error(irc);
		irc = UF_CURVE_edit_length(line, 1, -(tool_diameter/2+stock), 2, 
			UF_CURVE_LINEAR_SHAPE);
		if(irc) get_error(irc);
	} else 
	{
		start_point[0] = start_point[0]+distances[1]/2*directions[1][0];
		start_point[1] = start_point[1]+distances[1]/2*directions[1][1];
		start_point[2] = start_point[2]+distances[1]/2*directions[1][2];

		origin[0] = start_point[0]+z_max*directions[2][0];;
		origin[1] = start_point[1]+z_max*directions[2][1];;
		origin[2] = start_point[2]+z_max*directions[2][2];

		trochoida_id = create_trochoida(tolerance, start_point, 
			distances, mtx, tool_diameter, stock, step, 
			start_angle, list_edges, edge_count, start_edge, directions);
	}
	irc = UF_OBJ_delete_object(csys_id);
	if(irc) get_error(irc);	
	UF_MODL_delete_list(&edge_list);
	UF_MODL_delete_list(&face_list);
	free(list_edges);
	if(cut_type == 0) return line;
	else return trochoida_id;
}

void create_finish_boundary(double origin[3], tag_t face_id, 
							tag_t finish_operation)
{
	int ii,	entity_count = 0;
	int irc;
	UF_CAMBND_app_data_p_t app_data_list_finish[MAX_BND];
	UF_CAMBND_boundary_data_p_t bnd_data_finish;
	double matrix[9];
	char str[256] = "";
	tag_t entity_list_finish[MAX_BND], matrix_id;

	irc = UF_CSYS_ask_matrix_of_object(face_id, &matrix_id);
	if(irc) get_error(irc);
	irc = UF_CSYS_ask_matrix_values(matrix_id, matrix);
	if(irc) get_error(irc);

	for(ii=0; ii<num_bnd_curves; ii++)
	{
		entity_list_finish[entity_count] = bnd_edges[ii];
		app_data_list_finish[entity_count] = 
			create_and_init_finish_bnd_app_data();
		entity_count++;
	}
	bnd_data_finish = create_and_init_bnd_data(UF_CAM_boundary_type_open,
				0, origin, matrix, UF_CAM_material_side_undefined, 
				1, 1, 1, NULL);
	
	UF_CAMBND_append_bnd_from_curve(finish_operation, UF_CAM_part, 
		entity_count, entity_list_finish, bnd_data_finish, 
		app_data_list_finish);

	for (ii=0; ii<entity_count; ii++)
	{
		UF_free(app_data_list_finish[ii]);
	}
	UF_free(bnd_data_finish);
}

void create_boundary(tag_t boundary_curve, double origin[3], 
					 tag_t face_id, tag_t cam_oper)
{
	int i, irc;
	double matrix[9];
	tag_t entity_list[MAX_BND], matrix_id;
	UF_CAMBND_app_data_p_t app_data_list[MAX_BND];
	UF_CAMBND_boundary_data_p_t bnd_data;

	entity_list[0] = boundary_curve;
	app_data_list[0] = create_and_init_bnd_app_data();

	for(i=0; i < 1; i++)
	{
/* Ask matrix by object */
		irc = UF_CSYS_ask_matrix_of_object(face_id, &matrix_id);
		if(irc) get_error(irc);
		irc = UF_CSYS_ask_matrix_values(matrix_id, matrix);
		if(irc) get_error(irc);
/* Init boundary data */
		bnd_data = create_and_init_bnd_data(UF_CAM_boundary_type_open,
				0, origin, matrix, UF_CAM_material_side_undefined, 1, 1, 1, NULL);
/* Append curve boundaries. */
		irc = UF_CAMBND_append_bnd_from_curve(cam_oper, UF_CAM_part, 1, 
				&entity_list[0], bnd_data, app_data_list);
		if(irc) get_error(irc);
	}
/* Free memory */
	UF_free(app_data_list[0]);
	UF_free(bnd_data);
}

void create_boundary_edges(tag_t face_objects_sel, double origin[3])
{
	char str[256] = "";
	int edge_count;
	int i, j, k, irc, edge_type, norm_dir, type;
	tag_t edge_object, max_edge;
	tag_t *list_edges, start_edge;
	double arc_length, max_length = 0, tolerance;
	double start_point[3], end_point[3], dir[3];
	double point[3], box[6], radius, rad_data;

	uf_list_p_t edge_list;

	UF_MODL_ask_distance_tolerance(&tolerance);

	irc = UF_MODL_ask_face_data(face_objects_sel, &type, point, 
				dir, box, &radius, &rad_data, &norm_dir);
	if(irc) get_error(irc);

	irc = UF_MODL_ask_face_edges(face_objects_sel, &edge_list);
	if(irc) get_error(irc);
	irc = UF_MODL_ask_list_count(edge_list, &edge_count);
	if(irc) get_error(irc);
	list_edges = (tag_t *)malloc(sizeof(tag_t));
	j = 1;
	for(i = 0; i < edge_count; i++)
	{
		irc = UF_MODL_ask_list_item(edge_list, i, &edge_object);
		if(irc) get_error(irc);

		list_edges[i] = edge_object;
		list_edges = (tag_t *)realloc(list_edges, sizeof(tag_t)*(j+1));
		j++;
		irc = UF_MODL_ask_edge_type(edge_object, &edge_type);
		if(irc) get_error(irc);
		if(edge_type == UF_MODL_LINEAR_EDGE)
		{
			irc = UF_CURVE_ask_arc_length(edge_object, 0.0, 1.0, 
				UF_MODL_UNITS_PART, &arc_length);
			if(irc) get_error(irc);
			if(arc_length > max_length) 
			{
				max_length = arc_length;
				max_edge = edge_object;
			}
		}
	}
	ask_edge_start(max_edge, start_point);
	ask_edge_end(max_edge, end_point);
	start_edge = open_edge(list_edges, edge_count, tolerance);

/* Start find finish boundary */
	sort_end_to_end(list_edges, edge_count);
	j=1; k=0;
	bnd_edges = (tag_t *)malloc(sizeof(tag_t));
	for(i = 0; i < edge_count; i++)
	{
		if(list_edges[i] == start_edge)
		{
		}	
		else if(list_edges[i] != start_edge)
		{
			bnd_edges[k] = list_edges[i];
			bnd_edges = (tag_t *)realloc(bnd_edges, sizeof(tag_t)*(j+1));
			j++; 
			k++;
		}
	}
	sort_end_to_end(bnd_edges, k);
	num_bnd_curves = k;
/* End find finish boundary */
}

extern int tool_list_menu(int *response)
{
	int error_code = 0;
 
	if((error_code = UF_initialize()) != 0) return(0);

	if((error_code = UF_STYLER_create_dialog("tool_list.dlg",
           TOOL_cbs, TOOL_CB_COUNT, NULL, response)) != 0)
	{
		char fail_message[133];
		UF_get_fail_message(error_code, fail_message);
          UF_UI_set_status (fail_message);
          printf ( "%s\n", fail_message ); 
    }
    UF_terminate();                             
    return (error_code);
}


#include "mill_slots_service_function.h"
#include "mill_slots_styler_functions.h"
#include "tool_list_styler_functions.h"
